﻿using System;
using System.Threading;
using System.Diagnostics;


namespace WS2812_USB_UART
{

    internal delegate void IntervalElapsedEventHandler(Object sender, EventArgs e);
    
    internal partial class IntervalTimer : IDisposable
    {
        
        #region Public field

        public event IntervalElapsedEventHandler IntervalElapsed;
        
        public long MinEventZeit { get; private set; }
        public long AktEventZeit { get; private set; }
        public long MaxEventZeit { get; private set; }

        public long MaxStartIntervZeit { get; private set; }

        public long MinIntervZeit { get; private set; }
        public long AktIntervZeit { get; private set; }
        public long MaxIntervZeit { get; private set; }

        public int Zeituberschreitung { get; private set; }

        public double Auslastung { get; private set; }
        public double FPS { get; private set; }
        
        #endregion


        #region Private field

        private Thread ThreadW;
        private volatile bool StoppThreadW;

        private IntervalGenerator IntGen;

        private readonly long IntervalTime;

        #endregion


        #region Constructor

        public IntervalTimer(long intervalTime, IGJitter iGJitter)
        {
            IntervalTime = intervalTime;

            IntGen = new IntervalGenerator(intervalTime, iGJitter);

            ResetCounter();
        }

        #endregion


        #region Dispose

        private bool disposed;

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing) { Stop(); }

                IntGen.Dispose();

                disposed = true;
            }
        }   // Dispose


        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }


        ~IntervalTimer()
        {
            Dispose(false);
        }

        #endregion


        #region Start / Stop / Reset

        public void Start()
        {
            if (disposed)
                return;

            Stop();
            
            // Thread erzeugen / starten
            StoppThreadW = false;
            ThreadW = new Thread(ThreadWorker);
            ThreadW.IsBackground = true;
            //ThreadW.Priority = ThreadPriority.Highest;
            ThreadW.Start();
        }
        

        public void Stop()
        {
            if (ThreadW != null)
            {
                StoppThreadW = true;
                ThreadW.Join();
                ThreadW = null;
            }

            Auslastung = 0;
            FPS = 0;
        }

        public void ResetCounter()
        {
            MinEventZeit = 100000;
            AktEventZeit = 0;
            MaxEventZeit = 0;

            MaxStartIntervZeit = 0;

            MinIntervZeit = 100000;
            AktIntervZeit = 0;
            MaxIntervZeit = 0;

            Zeituberschreitung = 0;
        }

        #endregion


        #region Arbeitstier

        private void ThreadWorker()
        {
            AverageCalc AuslastungAv = new AverageCalc(5000);
            AverageCalc FpsAv = new AverageCalc(200);

            long ElapstTime;
            long startTime = 0;
            long spanTime;

            long OldFrameTick = Stopwatch.GetTimestamp();

            GC.Collect(2, GCCollectionMode.Forced);
            GC.WaitForPendingFinalizers();

            IntGen.ResetInterval();

            long gctimer = 0;

            bool ErsterZyklus = true;

            while (!StoppThreadW)
            {

                // Jede Sekunden eine GC Gen1 forcen
                long t = Stopwatch.GetTimestamp() * 1000 / Stopwatch.Frequency;
                if (t > gctimer + 1000)
                {
                    gctimer = t;
                    GC.Collect(1, GCCollectionMode.Forced);
                }
                

                if (!ErsterZyklus)
                {
                    ElapstTime = IntGen.ElapstTime;

                    AktIntervZeit = ElapstTime;

                    if (ElapstTime < MinIntervZeit) { MinIntervZeit = ElapstTime; }
                    if (ElapstTime > MaxIntervZeit) { MaxIntervZeit = ElapstTime; }
                    if (ElapstTime > IntervalTime) { Zeituberschreitung += 1; }

                    spanTime = ElapstTime - startTime;
                    Auslastung = AuslastungAv.NewValue(spanTime * 100 / (IntervalTime - (TimerResolution.CurrentResolution / 10)));
                }

                
                // Warten auf neuen Intervall
                IntGen.WaitForInterval();


                startTime = IntGen.ElapstTime;
                if (startTime > MaxStartIntervZeit) { MaxStartIntervZeit = startTime; }


                long stwtick = Stopwatch.GetTimestamp();
                double aktstwt = FpsAv.NewValue(stwtick - OldFrameTick);
                OldFrameTick = stwtick;
                FPS = Stopwatch.Frequency / aktstwt;


                ElapstTime = IntGen.ElapstTime;


                // Arbeit erledigen
                if (IntervalElapsed != null) { IntervalElapsed(this, EventArgs.Empty); }


                spanTime = IntGen.ElapstTime - ElapstTime;
                AktEventZeit = spanTime;
                if (spanTime < MinEventZeit) { MinEventZeit = spanTime; }
                if (spanTime > MaxEventZeit) { MaxEventZeit = spanTime; }

                ErsterZyklus = false;
            }

        }

        #endregion
  
    }


    #region Durchschnitt

    internal class AverageCalc
    {
        private long[] BufferAverage;
        private int iLength;
        private double dLength;
        private int InOutPointer;
        private int Start;

        public AverageCalc(int length)
        {
            BufferAverage = new long[length];
            dLength = (double)(iLength = length);
        }

        public double Average { get; private set; }

        private long Summe;

        public double NewValue(long newValue)
        {
            Summe -= BufferAverage[InOutPointer];
            Summe += newValue;
            BufferAverage[InOutPointer++] = newValue;

            if (InOutPointer >= iLength)
                InOutPointer = 0;

            if (Start < iLength)
            {
                Start++;
                Average = Summe / (double)Start;
            }
            else
                Average = Summe / dLength;

            return Average;
        }

        public void Reset()
        {
            Array.Clear(BufferAverage, 0, iLength);
            Summe = Start = 0;
            Average = 0.0;
        }

    }

    #endregion

}
